/*
 *	UUID+password-based authentication module for Vbox-VRDP
 *	Copyright © Jan Engelhardt <jengelh [at] medozas de>, 2009 - 2010
 *
 *	This file is part of vrdp-ldap. vrdp-ldap is free software; you can
 *	redistribute it and/or modify it under the terms of the GNU
 *	Lesser General Public License as published by the Free Software
 *	Foundation; version 3 of the License.
 */
#include <errno.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <libHX/defs.h>
#include <libHX/option.h>
#include <libHX/string.h>
#include <security/pam_appl.h>
#include <security/pam_modules.h>
#define LDAP_DEPRECATED 1
#include <ldap.h>
#include <VRDPAuth.h>

union guid {
	uint8_t raw[16];
	struct {
		uint32_t a;
		uint16_t b, c;
		uint64_t d;
	};
};

/**
 * struct muuid_state - container for all our belongings
 *
 * @uuid:	Machine UUID in text-form ("{8-4-4-4-12}")
 */
struct muuid_state {
	/* Configuration */
	char *ldap_uri, *bind_dn, *bind_pw, *base_dn;
	char *pam_service;
	unsigned int check_pam, check_uuid;

	/* Runtime */
	LDAP *conn;
	const char *user, *password;
	char uuid[39];
};

extern char *muuid_strquote(const char *, char **);	/* temporary */

static void muuid_conf_init(struct muuid_state *state)
{
	static const char *config_file = CONFIG_SYSCONFDIR "/vboxmuuid.conf";
	struct HXoption options_table[] = {
		{.ln = "CHECK_PAM", .type = HXTYPE_NONE, .ptr = &state->check_pam},
		{.ln = "CHECK_UUID", .type = HXTYPE_NONE, .ptr = &state->check_uuid},
		{.ln = "LDAP_URI", .type = HXTYPE_STRING, .ptr = &state->ldap_uri},
		{.ln = "BIND_DN", .type = HXTYPE_STRING, .ptr = &state->bind_dn},
		{.ln = "BIND_PW", .type = HXTYPE_STRING, .ptr = &state->bind_pw},
		{.ln = "BASE_DN", .type = HXTYPE_STRING, .ptr = &state->base_dn},
		{.ln = "PAM_SERVICE", .type = HXTYPE_STRING, .ptr = &state->pam_service},
		HXOPT_TABLEEND,
	};
	int ret;

	memset(state, 0, sizeof(*state));
	state->check_pam = state->check_uuid = true;
	ret = HX_shconfig(config_file, options_table);
	if (ret < 0)
		fprintf(stderr, "%s: could not read %s: %s\n",
		        __func__, config_file, strerror(-ret));
}

static void muuid_conf_exit(struct muuid_state *state)
{
	free(state->ldap_uri);
	free(state->bind_dn);
	free(state->bind_pw);
	free(state->base_dn);
	free(state->pam_service);
	memset(state, 0, sizeof(*state));
}

/**
 * muuid_ldap_open - connect to LDAP server
 *
 * Returns <=0 for failure (with error code), or 1 for success.
 */
static int muuid_ldap_open(struct muuid_state *state)
{
	int ret;

	if (state->ldap_uri == NULL)
		return -EINVAL;

	ret = ldap_initialize(&state->conn, state->ldap_uri);
	if (ret != LDAP_SUCCESS)
		return 0;

	ret = LDAP_VERSION3;
	ldap_set_option(state->conn, LDAP_OPT_PROTOCOL_VERSION, &ret);
	ldap_start_tls_s(state->conn, NULL, NULL);
	return ldap_simple_bind_s(state->conn, state->bind_dn,
	       state->bind_pw) == LDAP_SUCCESS;
}

/**
 * muuid_ldap_auth - do UUID verification
 *
 * Checks whether the given user is cleared for a particular machine
 * (match by existence of an attribute carrying the UUID).
 *
 * Returns <0 on error, 0 for auth denied, and >0 for auth granted.
 */
static int muuid_ldap_auth(struct muuid_state *state)
{
	LDAPMessage *result;
	char *p, *free_me = NULL;
	hxmc_t *filter;
	int ret;

	filter = HXmc_strinit("(&(objectClass=VBoxVRDP)(VBoxVRDPUUID=");
	if (filter == NULL)
		return -errno;
	p = muuid_strquote(state->uuid, &free_me);
	if (p == NULL)
		goto out;
	if (HXmc_strcat(&filter, p) == NULL)
		goto out;
	if (HXmc_strcat(&filter, ")(uid=") == NULL)
		goto out;
	p = muuid_strquote(state->user, &free_me);
	if (p == NULL)
		goto out;
	if (HXmc_strcat(&filter, p) == NULL)
		goto out;
	if (HXmc_strcat(&filter, "))") == NULL)
		goto out;

	ret = ldap_search_ext_s(state->conn, state->base_dn,
	      LDAP_SCOPE_SUBTREE, filter, NULL, true, NULL,
	      NULL, NULL, 1, &result);
	if (ret != LDAP_SUCCESS || result == NULL) {
		ret = 0;
	} else {
		if (ldap_first_entry(state->conn, result) != NULL)
			ret = 1;
		ldap_msgfree(result);
	}
 out:
	HXmc_free(filter);
	free(free_me);
	return ret;
}

static void muuid_ldap_close(struct muuid_state *state)
{
	ldap_unbind_ext(state->conn, NULL, NULL);
}

/**
 * muuid_ldap_run - do the LDAP UUID verification
 *
 * Returns <0 on error, 0 for auth denied, and >0 for auth granted.
 */
static int muuid_ldap_run(struct muuid_state *state)
{
	int ret;

	ret = muuid_ldap_open(state);
	if (ret <= 0) {
		fprintf(stderr, "muuid_ldap_open: %s\n", strerror(-ret));
		return ret;
	}

	ret = muuid_ldap_auth(state);
	if (ret < 0)
		fprintf(stderr, "muuid_ldap_auth: %s\n", strerror(-ret));
	else if (ret == 0)
		fprintf(stderr, "%s: denied: %s\n", __func__, state->user);
	else
		fprintf(stderr, "%s: granted: %s\n", __func__, state->user);
	muuid_ldap_close(state);
	return ret;
}

static int muuid_pam_conv(int num_msg, const struct pam_message **msg_ap,
    struct pam_response **res_ap, void *ptr)
{
	int j;

	for (j = 0; j < num_msg; ++j) {
		switch (msg_ap[j]->msg_style) {
		case PAM_PROMPT_ECHO_ON:
			/* username given to pam already */
			return PAM_CONV_ERR;
		case PAM_PROMPT_ECHO_OFF:
			res_ap[j] = malloc(sizeof(struct pam_response));
			if (res_ap[j] == NULL)
				goto free_up;
			res_ap[j]->resp_retcode = PAM_SUCCESS;
			res_ap[j]->resp = strdup(ptr);
			break;
		case PAM_TEXT_INFO:
			break;
		default:
			return PAM_CONV_ERR;
		}
	}
	return PAM_SUCCESS;

 free_up:
	for (; j >= 0; --j) {
		struct pam_response *r = res_ap[j];
		free(r->resp);
		free(r);
	}
	errno = ENOMEM;
	return PAM_CONV_ERR;
}

/**
 * muuid_pam_authenticate - do PAM authentication
 *
 * This verifies existence of user and matches passwords.
 * Returns a PAM error code.
 */
static int muuid_pam_auth(const struct muuid_state *state)
{
	const struct pam_conv conv = {
		.conv        = muuid_pam_conv,
		.appdata_ptr = (void *)state->password,
	};
	pam_handle_t *pamh;
	int ret;

	ret = pam_start(state->pam_service, state->user, &conv, &pamh);
	if (ret != PAM_SUCCESS) {
		fprintf(stderr, "%s: pam_start: %s\n",
		        __func__, pam_strerror(pamh, ret));
		return ret;
	}

	ret = pam_authenticate(pamh, 0);
	if (ret != PAM_SUCCESS)
		fprintf(stderr, "%s: pam_authenticate: %s\n",
		        __func__, pam_strerror(pamh, ret));
	pam_end(pamh, ret);
	return ret;
}

/**
 * Converts a 128-bit UUID in binary encoding into the 36-byte text
 * representation.
 */
static void muuid_b2t(char *t, const uint8_t *b)
{
	static const char hex[] = "0123456789abcdef";
	unsigned int i, j;
	union guid guid;

	memcpy(&guid, b, sizeof(guid));
	guid.a = htonl(guid.a);
	guid.b = htons(guid.b);
	guid.c = htons(guid.c);
	b = guid.raw;

	*t++ = '{';
	for (i = 0; i < 4; ++i) {
		*t++ = hex[(*b >> 4) & 0x0F];
		*t++ = hex[*b++ & 0x0F];
	}
	for (j = 0; j < 3; ++j) {
		*t++ = '-';
		for (i = 0; i < 2; ++i) {
			*t++ = hex[(*b >> 4) & 0x0F];
			*t++ = hex[*b++ & 0x0F];
		}
	}
	*t++ = '-';
	for (i = 0; i < 6; ++i) {
		*t++ = hex[(*b >> 4) & 0x0F];
		*t++ = hex[*b++ & 0x0F];
	}
	*t++ = '}';
	*t++ = '\0';
}

static bool muuid_run(const char *user, const char *password, uint8_t *b_uuid)
{
	struct muuid_state state;
	bool succ = true;

	muuid_conf_init(&state);
	state.user     = user;
	state.password = password;
	if (b_uuid != NULL)
		muuid_b2t(state.uuid, b_uuid);
	fprintf(stderr, "%s: user=%s uuid=%s\n", __func__, user, state.uuid);

	if (state.check_uuid && muuid_ldap_run(&state) <= 0)
		succ = false;
	if (state.check_pam && muuid_pam_auth(&state) != PAM_SUCCESS)
		succ = false;

	muuid_conf_exit(&state);
	return succ;
}

/* ugly code below here */

extern VRDPAuthResult VRDPAUTHCALL VRDPAuth(PVRDPAUTHUUID,
	VRDPAuthGuestJudgement, const char *, const char *, const char *);

VRDPAuthResult VRDPAUTHCALL
VRDPAuth(PVRDPAUTHUUID uuid, VRDPAuthGuestJudgement guest_judgement,
         const char *user, const char *password, const char *domain)
{
	return muuid_run(user, password, *uuid) ?
	       VRDPAuthAccessGranted : VRDPAuthAccessDenied;
}
